import re

import boto3

console = boto3.client('apigateway')


def fetch_api_by_name(api_name):
    """ Fetch an api record by its name """
    api_records = console.get_rest_apis()['items']
    matches = filter(lambda x: x['name'] == api_name, api_records)
    if not matches:
        return None
    return matches[0]


def fetch_resources_by_api(api_id):
    """ Fetch all resources for an apigateway rest api """
    return console.get_resources(restApiId=api_id)['items']


def fetch_resource(api_id, resource_id):
    """ Fetch extra metadata for this particular resource """
    return console.get_resource(restApiId=api_id, resourceId=resource_id)


def fetch_method(api_id, resource_id, verb):
    """ Fetch extra metadata for this particular method """
    return console.get_method(
            restApiId=api_id,
            resourceId=resource_id,
            httpMethod=verb)


def lambda_arn_from_method_integration_arn(int_arn):
    pattern = '.*functions/(.*)/invocations'
    matches = re.match(pattern, int_arn)
    return matches.group(1)


def lambda_name_from_lambda_arn(lambda_arn):
    return lambda_arn.split(':')[-1]


def fetch_root_resource_for_api(api_id):
    records = console.get_resources(restApiId=api_id)['items']
    for resource in records:
        if resource['path'] == '/':
            return resource



import unittest
from psychic_disco import util
import os
import shutil


class TestUtil(unittest.TestCase):
    def test_shell(self):
        util.shell(["touch", "poop.txt"])
        self.assertTrue(os.path.exists("poop.txt"))
        os.remove("poop.txt")

    def test_cp(self):
        with open("poop_src.txt", "w") as f:
            f.write("poop")
        util.cp("poop_src.txt", "poop_dest.txt")
        self.assertTrue(os.path.exists("poop_dest.txt"))
        os.remove("poop_src.txt")
        os.remove("poop_dest.txt")

    def test_zip(self):
        shutil.rmtree("poop", True)
        os.mkdir("poop")
        with open("poop/poop_1.txt", "w") as f:
            f.write("poop1")
        with open("poop/poop_2.txt", "w") as f:
            f.write("poop2")
        util.zip("poop", "poop.zip")
        self.assertTrue(os.path.exists("poop.zip"))
        shutil.rmtree("poop", True)
        os.remove("poop.zip")

    def test_py_files_in_directory(self):
        shutil.rmtree("poop", True)
        os.mkdir("poop")
        with open("poop/poop_1.py", "w") as f:
            f.write("poop1")
        with open("poop/poop_2.py", "w") as f:
            f.write("poop2")
        with open("poop/poop_2.txt", "w") as f:
            f.write("poop2")
        py_files = []
        for py_file in util.py_files_in_directory("poop"):
            py_files.append(py_file)
        self.assertEqual(2, len(py_files))
        self.assertIn("poop/poop_2.py", py_files)
        self.assertIn("poop/poop_1.py", py_files)
        shutil.rmtree("poop", True)

if __name__ == "__main__":
    unittest.main()



import unittest
from psychic_disco import lambda_function
from psychic_disco.lambda_function import LambdaFunction


class TestLambdaFunction(unittest.TestCase):
    def test_declared_functions(self):
        f = LambdaFunction("listdir", "os")
        self.assertIn(f, lambda_function.declared_functions())

    def test_proper_name(self):
        f = LambdaFunction("listdir", "os")
        self.assertEqual(f.proper_name, "os-listdir")

if __name__ == "__main__":
    unittest.main()



import unittest

from psychic_disco.rest_method import RestMethod
from psychic_disco.rest_resource import RestResource
from psychic_disco.rest_api import RestApi


class TestRestMethod(unittest.TestCase):
    def test_fetch_from_aws(self):
        method = RestMethod('post')
        method.resource = RestResource('/dummies')
        method.resource.aws_id = "123resourceid"
        method.resource.api = RestApi.find("dummy-manager")
        method.api = method.resource.api

        self.assertIsNone(method.lambda_name)
        method.fetch_from_aws()
        self.assertEqual(
                "psychic_disco-dummy_lambda_function",
                method.lambda_name)



import unittest

from psychic_disco.rest_resource import RestResource
from psychic_disco.rest_api import RestApi


class TestRestResource(unittest.TestCase):
    def test_rest_resource_not_exists(self):
        resource = RestResource('/dummies')
        self.assertFalse(resource.exists_in_aws)

    def test_rest_resource_exists(self):
        api = RestApi.find("dummy-manager")
        resource = RestResource('/dummies')
        resource.aws_id = "123resourceid"
        resource.api = api
        self.assertTrue(resource.exists_in_aws)

    def test_rest_resource_fetch_from_aws(self):
        api = RestApi.find("dummy-manager")
        resource = RestResource('/dummies')
        resource.aws_id = "123resourceid"
        resource.api = api

        self.assertEqual(0, len(resource.methods))
        resource.fetch_from_aws()
        self.assertEqual(1, len(resource.methods))

    def test_rest_resource_path_part(self):
        api = RestApi.find("dummy-manager")
        resource = RestResource('/dummies/by_name')
        resource.api = api
        self.assertEqual(resource.path_part, "by_name")



import os
import json


class FakeApiGatewayConsole(object):
    def __init__(self):
        self.apis = dict()
        self.basepath = os.path.dirname(os.path.abspath(__file__))

    def _open_json(self, filename):
        path = os.path.join(self.basepath, filename)
        with open(path) as f:
            return json.loads(f.read())

    def get_rest_apis(self):
        return self._open_json("get_rest_apis.json")

    def get_resources(self, **kwargs):
        return self._open_json("get-resources.json")

    def get_resource(self, **kwargs):
        return self._open_json("get-resource.json")

    def get_method(self, **kwargs):
        return self._open_json("get-method.json")



import psychic_disco.api_gateway_config as agc
from psychic_disco.rest_method import RestMethod


class RestResource(object):
    def __init__(self, path):
        self.path = path
        self.aws_id = None
        self.api = None
        self.methods = dict()

    @property
    def exists_in_aws(self):
        return (self.aws_id is not None and self.api is not None)

    def fetch_from_aws(self):
        if not self.exists_in_aws:
            raise Exception("Not enough info to query AWS for Rest Resource")
        record = agc.fetch_resource(self.api.aws_id, self.aws_id)
        if 'resourceMethods' in record:
            for verb in record['resourceMethods']:
                if verb not in self.methods:
                    self.methods[verb] = RestMethod(verb)
                method = self.methods[verb]
                method.resource = self
                method.api = self.api
                method.fetch_from_aws()

    @property
    def parent(self):
        if self.path == '/':
            return None
        else:
            parent_path = '/'.join(self.path.split('/')[0:-1]) or '/'
            return self.api.declare_resource(parent_path)

    @property
    def path_part(self):
        if self.parent:
            return self.path.replace(self.parent.path + "/", '', 1)
        else:
            return self.path

    def deploy(self):
        print "\tDeploying Resource: %s" % self.path
        if not self.exists_in_aws:
            self.parent.deploy()
            response = agc.console.create_resource(
                    restApiId=self.api.aws_id,
                    parentId=self.parent.aws_id,
                    pathPart=self.path_part)
            self.aws_id = response['id']
        for verb in self.methods:
            method = self.methods[verb]
            method.deploy()

    def declare_method(self, verb):
        if verb not in self.methods:
            m = RestMethod(verb)
            m.resource = self
            m.api = self.api
            self.methods[verb] = m
        return self.methods[verb]

    def declare_route(self, verb, lambda_function):
        m = self.declare_method(verb)
        m.declare_route(lambda_function)



""" Provides a simple convention for defining cli arguments and subcommands """
import argparse
import inspect

parser = argparse.ArgumentParser(prog='psychic_disco')
subparsers = parser.add_subparsers(help='sub-command help')
args = None


def _apply_args_to_func(global_args, func):
    """ Unpacks the argparse Namespace object and applies its
    contents as normal arguments to the function func """
    global_args = vars(global_args)
    local_args = dict()
    for argument in inspect.getargspec(func).args:
        local_args[argument] = global_args[argument]
    return func(**local_args)


def subcommand(f):
    """ Define a subcommand for the psychic_disco function
    Expects functions to supply a list of their arguments
    Converts these to arg_parse arguments, and transforms
    the argparse Namespace object so that commands can be
    invoked as though they are normal functions (which they are)"""
    sp = subparsers.add_parser(f.__name__, help=f.__doc__)
    for argument in inspect.getargspec(f).args:
        sp.add_argument(argument)
    sp.set_defaults(func=lambda args: _apply_args_to_func(args, f))
    return f


def parse_args():
    """ Parse command line arguments and assign contents to
    `cli.args` variable"""
    global args
    args = parser.parse_args()
    return args


def argument(*args, **kwargs):
    """ Set up a new argument for the psychic_disco program.
    Follows same conventions as argparse's add_argument """
    parser.add_argument(*args, **kwargs)



import psychic_disco
import psychic_disco.api_gateway_config as agc


class RestMethod(object):
    def __init__(self, verb):
        self.verb = verb
        self.resource = None
        self.api = None
        self.lambda_name = None
        self.lambda_function = None

    def declare_route(self, lf):
        self.lambda_function = lf
        self.lambda_name = lf.proper_name

    def fetch_from_aws(self):
        if (self.resource is None or self.api is None):
            msg = """I cannot possibly question AWS
            without knowing my resource and api ids"""
            raise Exception(msg)
        record = agc.fetch_method(
                self.api.aws_id,
                self.resource.aws_id,
                self.verb)
        arn = record['methodIntegration']['uri']
        lambda_arn = agc.lambda_arn_from_method_integration_arn(arn)
        self.lambda_name = agc.lambda_name_from_lambda_arn(lambda_arn)

    @property
    def exists_in_aws(self):
        try:
            agc.console.get_method(
                    restApiId=self.api.aws_id,
                    resourceId=self.resource.aws_id,
                    httpMethod=self.verb)
            return True
        except:
            return False

    @property
    def integration_uri(self):
        pattern = "arn:aws:apigateway:%s:lambda:path/2015-03-31/functions/%s/invocations"  # noqa
        region = psychic_disco.config.default_aws_region
        lambda_arn = self.lambda_function.arn
        return pattern % (region, lambda_arn)

    def deploy(self):
        print "\t\tDeploying Method: %s" % self.verb
        should_put_integration = False
        if not self.exists_in_aws:
            should_put_integration = True
            agc.console.put_method(
                    restApiId=self.api.aws_id,
                    resourceId=self.resource.aws_id,
                    httpMethod=self.verb,
                    authorizationType='NONE')
        if self.lambda_function:
            self.lambda_function.deploy()
        if should_put_integration:
            agc.console.put_integration(
                    restApiId=self.api.aws_id,
                    resourceId=self.resource.aws_id,
                    httpMethod=self.verb,
                    type='AWS',
                    integrationHttpMethod=self.verb,
                    uri=self.integration_uri)



import subprocess
import os


def shell(commands):
    print("==> %s" % ' '.join(commands))
    subprocess.call(commands)


def cp(src, dest):
    shell(['cp', src, dest])


def zip(directory, zip_file):
    original_dir = os.getcwd()
    explicit_zip_path = os.path.join(original_dir, zip_file)
    os.chdir(directory)
    shell(['zip', '-r',  explicit_zip_path, '.', '-i', '*'])
    os.chdir(original_dir)


def py_files_in_directory(directory):
    for (dirpath, dirnames, filenames) in os.walk(directory):
        for dirname in dirnames:
            if dirname.startswith("."):
                # Don't visit hidden directories
                dirnames.remove(dirname)
        for filename in filenames:
            # Don't mess with hidden files
            if filename.endswith(".py") and not filename.startswith("."):
                yield os.path.join(dirpath, filename)



#!/usr/bin/env python
import unittest
import psychic_disco
import example_package  # noqa


class TestApi(unittest.TestCase):
    def test_entry_points(self):
        expected = "psychic_disco.tests.example_package.handler"
        entry_point = psychic_disco.entry_points()[0]
        actual = entry_point.full_name
        self.assertEqual(expected, actual)

    def test_api_calls(self):
        expected = "psychic_disco.tests.example_package.hova"
        api_object = psychic_disco.Api["GET /that/dirt/off/your/shoulder"]
        actual = api_object.full_name
        self.assertEqual(expected, actual)

if __name__ == "__main__":
    unittest.main()



import unittest

from psychic_disco.rest_api import RestApi


class TestRestApi(unittest.TestCase):
    def test_find(self):
        api = RestApi.find("dummy-manager")
        self.assertEqual("123apiid", api.aws_id)
        self.assertTrue(api.exists_in_aws)
        self.assertEqual(2, len(api.resources))
        self.assertIn('/', api.resources)
        self.assertIn('/dummies', api.resources)

    def test_find_caches(self):
        api1 = RestApi.find("dummy-manager")
        api2 = RestApi.find("dummy-manager")
        self.assertEqual(api1, api2)

    def test_find_creates_nonexistent(self):
        api1 = RestApi.find("dummy-manager-nonexistent")
        self.assertIsNotNone(api1)



import psychic_disco


@psychic_disco.lambda_entry_point
def handler(event, context):
    print event


@psychic_disco.api_method("GET", "/that/dirt/off/your/shoulder")
def hova(event, context):
    print event



import psychic_disco.api_gateway_config as agc
from psychic_disco.rest_resource import RestResource


class RestApi(object):
    """ Model of an AWS API Gateway Rest API object. """

    apis_by_name = dict()
    """ Cache of API objects to ensure that there is
    a singleton for each actual API by name """

    def __init__(self, name):
        """ Not intended for external use. Call
        RestApi.find() method """
        self.name = name
        self.aws_id = None
        self.resources = dict()

    @classmethod
    def find(cls, api_name):
        """ Find or create an API model object by name """
        if api_name in cls.apis_by_name:
            return cls.apis_by_name[api_name]
        api = cls(api_name)
        api._fetch_from_aws()
        if api.exists_in_aws:
            api._fetch_resources()
        cls.apis_by_name[api_name] = api
        return api

    @property
    def root_resource(self):
        root = self.declare_resource('/')
        if not root.exists_in_aws:
            record = agc.fetch_root_resource_for_api(self.aws_id)
            root.aws_id = record['id']
        return root

    def _fetch_from_aws(self):
        record = agc.fetch_api_by_name(self.name)
        if record:
            self.aws_id = record['id']

    @property
    def exists_in_aws(self):
        return (self.aws_id is not None)

    def _fetch_resources(self):
        for record in agc.fetch_resources_by_api(self.aws_id):
            path = record['path']
            if path not in self.resources:
                self.resources[path] = RestResource(path)
            resource = self.resources[path]
            resource.aws_id = record['id']
            resource.api = self
            resource.fetch_from_aws()

    def deploy(self):
        print "Deploying API: %s" % self.name
        self._fetch_from_aws()
        if not self.exists_in_aws:
            agc.console.create_rest_api(name=self.name)
            self._fetch_from_aws()
        self.root_resource.deploy()  # It's special
        for path in sorted(self.resources.keys()):
            resource = self.resources[path]
            resource.deploy()
        agc.console.create_deployment(
                restApiId=self.aws_id,
                stageName='master')

    def declare_route(self, verb, path, lambda_function):
        r = self.declare_resource(path)
        r.declare_route(verb, lambda_function)

    def declare_resource(self, path):
        if path not in self.resources:
            r = RestResource(path)
            r.api = self
            self.resources[path] = r
        return self.resources[path]



import unittest

import psychic_disco.api_gateway_config as agc
from .fake_gateway import FakeApiGatewayConsole


class TestApiGatewayConfig(unittest.TestCase):
    def setUp(self):
        agc.console = FakeApiGatewayConsole()

    def test_fetch_api_by_name(self):
        # agc.console.apis["good_api"] = RestApi.find("good_api")
        actual = agc.fetch_api_by_name("dummy-manager")
        self.assertEqual("123apiid", actual['id'])

    def test_fetch_wrong_api_name(self):
        actual = agc.fetch_api_by_name("an api has no name")
        self.assertIsNone(actual)

    def test_fetch_resources_by_api(self):
        actual = agc.fetch_resources_by_api("123apiid")
        self.assertEqual(2, len(actual))
        self.assertEqual("/", actual[0]['path'])
        self.assertEqual("/dummies", actual[1]['path'])

    def test_fetch_resource(self):
        actual = agc.fetch_resource("123apiid", "123resourceid")
        self.assertEqual({}, actual['resourceMethods']['post'])

    def test_fetch_method(self):
        actual = agc.fetch_method("123apiid", "123resourceid", "post")
        uri = actual['methodIntegration']['uri']
        expected_ending = "psychic_disco-dummy_lambda_function/invocations"
        self.assertTrue(uri.endswith(expected_ending))

    def test_lambda_arn_from_method_integration_arn(self):
        method = agc.console._open_json("get-method.json")
        uri = method['methodIntegration']['uri']
        expected_arn = "arn:aws:lambda:us-east-1:123456789012:function:psychic_disco-dummy_lambda_function"  # noqa
        actual_arn = agc.lambda_arn_from_method_integration_arn(uri)
        self.assertEqual(expected_arn, actual_arn)

    def test_lambda_name_from_lambda_arn(self):
        arn = "arn:aws:lambda:us-east-1:123456789012:function:psychic_disco-dummy_lambda_function"  # noqa
        expected_name = "psychic_disco-dummy_lambda_function"
        actual_name = agc.lambda_name_from_lambda_arn(arn)
        self.assertEqual(expected_name, actual_name)


if __name__ == "__main__":
    unittest.main()
